/*
 **  curves.cpp
 **
 **  Simple curve drawing program.
 **
 **  The following keyboard commands are used to control the
 **  program:
 **
 **    q - Quit the program
 **    c - Clear the screen
 **    b - Cycle through Bezier (default), interpolating and B-spline curves
 **    g - Cycle between none, G^1 and C^1 continuity
 **
 */

#include <iostream>
#include <algorithm>
#include "./common.h"
#include "./vec.h"

using namespace std;

typedef enum {
  BEZIER = 0,
  INTERPOLATED = 1,
  BSPLINE = 2
} curveType;

curveType cur_type = BEZIER;
int line_segs = 30;
int continuity = 0;

void keyboard(unsigned char key, int x, int y);
void computeMatrix(curveType type, float m[4][4]);
void vmult(float m[4][4], float v[4][3], float r[4][3]);

/* Colors to draw them in */
GLfloat colors[][3] = {
  { 1.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 }
};

#define MAX_CPTS  25            /* Fixed maximum number of control points */

GLfloat cpts[MAX_CPTS][3];
GLfloat cpts_dud[MAX_CPTS][3];
int ncpts = 0;
int ncpts_dud = 0;

static int width = 500, height = 500;           /* Window width and height */

/* Matrix stuff */

/* This routine multiplies two 4 x 4 matrices. */

/* This routine multiplies a 4 x 4 matrix with a vector of 4 points. */
void vmult(float m[4][4], float v[4][3], float r[4][3])
{
  int i, j, k;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 3; j++)
      for (k = 0, r[i][j] = 0.0; k < 4; k++)
        r[i][j] += m[i][k] * v[k][j];
}


/* Interpolating to Bezier matrix */
static float minterp[4][4] = 
{
  { 1.0, 0.0, 0.0, 0.0 },
  { -5.0/6.0, 3.0, -3.0/2.0, 1.0/3.0 },
  { 1.0/3.0, -3.0/2.0, 3.0, -5.0/6.0 },
  { 0.0, 0.0, 0.0, 1.0 }, 
};

/* B-spline to Bezier matrix */
static float mbspline[4][4] = 
{
  { 1.0/6.0, 4.0/6.0, 1.0/6.0, 0.0 },
  { 0.0, 4.0/6.0, 2.0/6.0, 0.0 },
  { 0.0, 2.0/6.0, 4.0/6.0, 0.0 },
  { 0.0, 1.0/6.0, 4.0/6.0, 1.0/6.0 },
};

static float midentity[4][4] =
{
  { 1.0, 0.0, 0.0, 0.0},
  { 0.0, 1.0, 0.0, 0.0},
  { 0.0, 0.0, 1.0, 0.0},
  { 0.0, 0.0, 0.0, 1.0}
};

/* Calculate the matrix used to transform the control points */
void computeMatrix(curveType type, float m[4][4]) {
  int i, j;

  switch (type) {
    case BEZIER:
      /* Identity matrix */
      for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
          m[i][j] = midentity[i][j];
      break;
    case INTERPOLATED:
      for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
          m[i][j] = minterp[i][j];
      break;
    case BSPLINE:
      for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
          m[i][j] = mbspline[i][j];
      break;
  }
}

// Draw the indicated curves using the current control points.
void drawCurves(curveType type) {
  // Set the control point computation matrix and the step size.
  float m[4][4];
  computeMatrix(type, m);

  int step;
  if(type == BSPLINE) step = 1;
  else step = 3;

  glColor3fv(colors[type]);

  // Draw the curves
  int i = 0;
  while (i + 3 < ncpts) {
    // Calculate the appropriate control points
    GLfloat newcpts[4][3];
    vmult(m, &cpts[i], newcpts);

    // Draw the curve using OpenGL evaluators
    glMap1f(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, 4, &newcpts[0][0]);
    glMapGrid1f(line_segs, 0.0, 1.0);
    glEvalMesh1(GL_LINE, 0, line_segs);

    // Advance to the next segment
    i += step;
  }
  glFlush();
}

/* This routine displays the control points */
static void display(void) {
  int i;
  glClear(GL_COLOR_BUFFER_BIT);

  glColor3f(0.0, 0.0, 0.0);
  glPointSize(5.0);
  glBegin(GL_POINTS);
  for (i = 0; i < ncpts; i++)
    glVertex3fv(cpts[i]);
  glEnd();

  drawCurves(cur_type);

  glColor3f(0.5, 0.5, 0.5);
  glBegin(GL_LINE_STRIP);
  for (int i = 0; i < ncpts; ++i) {
    glVertex3fv(cpts[i]);
  }
  glEnd();

  glFlush();
}

void set(const Vec2f& v, int i) {
  cpts[i][0] = v[0];
  cpts[i][1] = v[1];
}

int cur_pt = -1;
void mouseMotion(int x, int y) {
  if (cur_pt != -1) {
    // Translate to our coordinate system
    const float wx = (2.0 * x) / (float)(width - 1) - 1.0;
    const float wy = (2.0 * (height - 1 - y)) / (float)(height - 1) - 1.0;
    const Vec2f w(wx, wy);
    const int i = cur_pt;
    if (continuity == 0 || cur_pt < 2 || cur_pt >= ncpts-2) {
      set(w, cur_pt);
    } else if (i%3 == 2) {
      // Moving a, change c
      set(w, cur_pt);
      Vec2f a(cpts[cur_pt]);
      Vec2f b(cpts[cur_pt+1]);
      Vec2f c(cpts[cur_pt+2]);
      c = (continuity==1)?(b+(b-a).unit()*(c-b).norm()) : (b+(b-a));
      set(c, cur_pt+2);
    } else if (i%3 == 0) {
      // Moving b, change a and c
      Vec2f a(cpts[cur_pt-1]);
      Vec2f b(cpts[cur_pt]);
      Vec2f c(cpts[cur_pt+1]);
      const Vec2f da = a-b;
      const Vec2f dc = c-b;
      b = Vec2f(wx, wy);
      set(b+da, cur_pt-1);
      set(b, cur_pt);
      set(b+dc, cur_pt+1);
    } else if (i%3 == 1) {
      // Moving c, change a
      set(w, cur_pt);
      Vec2f a(cpts[cur_pt-2]);
      Vec2f b(cpts[cur_pt-1]);
      Vec2f c(cpts[cur_pt]);
      a = (continuity==1)?(b+(b-c).unit()*(a-b).norm()) : (b+(b-c));
      set(a, cur_pt-2);
    }
    glutPostRedisplay();
  }
}

const float BUFFER = 0.005;

// This routine inputs new control points
void mouse(int button, int state, int x, int y) {
  // Translate to our coordinate system
  const float wx = (2.0 * x) / (float)(width - 1) - 1.0;
  const float wy = (2.0 * (height - 1 - y)) / (float)(height - 1) - 1.0;

  cur_pt = -1;

  // We are only interested in left clicks
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    // See if we're close to an existing point
    for (int i = 0; cur_pt == -1 && i < ncpts; ++i) {
      const float dx = wx-cpts[i][0];
      const float dy = wy-cpts[i][1];
      if (dx*dx+dy*dy < BUFFER) {
        cur_pt = i;
      }
    }

    if (cur_pt == -1) {
      // See if we have room for any more control points
      if (ncpts == MAX_CPTS)
        return;

      // Save the point
      cpts[ncpts][0] = wx;
      cpts[ncpts][1] = wy;
      cpts[ncpts][2] = 0.0;
      ncpts++;
    }
    glutPostRedisplay();
  } else if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) {
    // /* Translate back to our coordinate system */
    // wx = (2.0 * x) / (float)(width - 1) - 1.0;
    // wy = (2.0 * (height - 1 - y)) / (float)(height - 1) - 1.0;

    /* See if we have room for any more control points */
    if (ncpts_dud == MAX_CPTS)
      return;

    /* Save the point */
    cpts_dud[ncpts_dud][0] = wx;
    cpts_dud[ncpts_dud][1] = wy;
    cpts_dud[ncpts_dud][2] = 0.0;
    ncpts_dud++;

    /* Draw the point */
    glColor3f(1.0, 0.0, 0.0);
    glPointSize(5.0);
    glBegin(GL_POINTS);
    glVertex3f(wx, wy, 0.0);
    glEnd();

    glFlush();
  }
}


/* This routine handles keystroke commands */
void keyboard(unsigned char key, int x, int y) {
  static curveType lasttype = BEZIER;

  switch (key) {
    case 'q': case 'Q':
      exit(0);
      break;
    case 'c': case 'C':
      ncpts = 0;
      ncpts_dud = 0;
      glutPostRedisplay();
      break;
    case 'g': {
      continuity = (continuity+1)%3;
      if (continuity == 1) {
        for (int i = 3; i < ncpts-1; i+=3) {
          Vec2f a(cpts[i-1]);
          Vec2f b(cpts[i]);
          Vec2f c(cpts[i+1]);
          c = b + (b-a).unit() * (c-b).norm();
          cpts[i+1][0] = c[0];
          cpts[i+1][1] = c[1];
        }
      } else if (continuity == 2) {
        for (int i = 3; i < ncpts-1; i+=3) {
          Vec2f a(cpts[i-1]);
          Vec2f b(cpts[i]);
          Vec2f c(cpts[i+1]);
          c = b + (b-a);
          cpts[i+1][0] = c[0];
          cpts[i+1][1] = c[1];
        }
      }
      glutPostRedisplay();
      break;
    }
    // case 'd': case 'D':
    //   ncpts_dud = 0;
    //   glutPostRedisplay();
    //   break;
    case 'd':
      line_segs = std::max(line_segs-1, 1);
      glutPostRedisplay();
      break;
    case 'f':
      ++line_segs;
      glutPostRedisplay();
      break;
    case 'e': case 'E':
      glutPostRedisplay();
      break;
    case 'b': case 'B':
      drawCurves(BEZIER);
      lasttype = BEZIER; 
      cur_type = (curveType)((cur_type+1)%3);
      glutPostRedisplay();
      break;
    case 'i': case 'I':
      drawCurves(INTERPOLATED);
      lasttype = INTERPOLATED; 
      break;
    case 's': case 'S':
      drawCurves(BSPLINE);
      lasttype = BSPLINE; 
      break;
  }
}


/* This routine handles window resizes */
void reshape(int w, int h) {
  width = w;
  height = h;

  /* Set the transformations */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glViewport(0, 0, w, h);
}


main(int argc, char **argv) {
  cout << endl;
  cout << "Left-click to add control points.  Left-click and drag to " << endl;
  cout << "move a control point." << endl << endl;
  cout << "Key commands:" << endl;
  cout << "  c - Clear the screen" << endl;
  cout << "  b - Cycle through Bezier (default), interpolating and B-spline curves" << endl;
  cout << "  g - Cycle between none, G^1 and C^1 continuity" << endl;
  cout << "  q - Quit the program" << endl;

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB);
  glutInitWindowSize(width, height);
  glutCreateWindow("curves");

  // Register the callbacks
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutMotionFunc(mouseMotion);
  glutKeyboardFunc(keyboard);
  glutReshapeFunc(reshape);
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glEnable(GL_MAP1_VERTEX_3);

  glutMainLoop();
}
